import re
import string

from dataclasses import dataclass, field

_RE_DATETIME = re.compile(r"(\d{4}-\d{2}-\d{2})(?:(@\d{2}:\d{2}))?")

@dataclass
class Tx:
  date    : str                   = ""
  time    : str|None              = None
  summary : str                   = ""
  journals: list[tuple[str, int]] = field(default_factory=lambda: [])

def parse(stream):
  parts    = []
  end      = False
  line_num = 0
  while not end:
    line_num += 1

    line = stream.readline()
    end  = ("" == line)

    if end or not line[0].isspace():
      if tx := _make_tx(parts):
        yield tx
      parts = []

    comment = line.find("#")
    if comment >= 0:
      line = line[0:comment]

    line = line.strip()
    if line != "":
      parts.append((line_num, line))

def _make_tx(lines: list[tuple[int, str]]) -> dict:
  if len(lines) < 2:
    return None

  tx = Tx()

  line1 = lines[0][1].split(maxsplit=1)
  if m := re.fullmatch(_RE_DATETIME, line1[0]):
    tx.date = m.group(1)
    tx.time = m.group(2)
  else:
    raise Exception(f"invalid datetime format: '{line1[0]}' ({lines[0][1]})")

  if len(line1) == 2:
    tx.summary = line1[1]

  for line in lines[1:]:
    tokens = line[1].split()
    if len(tokens) != 2:
      raise Exception("invalid journal: '{line[1]}' (line[0])")
    t, name = _parse_entry_name(tokens[0])
    tx.journals.append((t, name, int(tokens[1])))
  return tx

def _parse_entry_name(name: str) -> tuple[str, str]:
  tokens = name.split("/", maxsplit=1)
  if len(tokens) != 2:
    raise Exception(f"invalid entry name: {name}")

  if "asset" == tokens[0]:
    t = "A"
  elif "debt" == tokens[0]:
    t = "D"
  elif "net" == tokens[0]:
    t = "N"
  elif "expense" == tokens[0]:
    t = "E"
  elif "revenue" == tokens[0]:
    t = "R"
  else:
    raise Exception(f"invalid entry type: {tokens[0]}")

  return (t, tokens[1])
